"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("reflect-metadata");
let __optional = Symbol('optional');
let __keys = Symbol('keys');
let __mapping = Symbol('mapping');
let __union = Symbol('union');
let __code = Symbol('code');
let __gt = Symbol('gt');
let __lt = Symbol('lt');
let __eq = Symbol('eq');
let __ne = Symbol('ne');
let __gte = Symbol('gte');
let __lte = Symbol('lte');
let __default = Symbol('default');
let __passthrough = Symbol('passthrough');
let __integer = Symbol('integer');
let __generic = Symbol('generic');
let __array = Symbol('array');
class JSONObject extends Object {
    newValue(value, design_type, key) {
        if (typeof value == 'object') {
            if (Array.isArray(value)) {
                let array_meta = Reflect.hasMetadata(__array, this, key) ? Reflect.getMetadata(__array, this, key) : null;
                if (array_meta && value.length) {
                    let ret = Array();
                    for (let i = 0; i < value.length; ++i) {
                        let e = value[i];
                        let new_e;
                        if (typeof e == 'object') {
                            if (Array.isArray(e)) { // (at least) 2 dim array, throw an unsupported error
                                throw new TypeError(`${this.constructor.name}.${key}: array type cannot be array, consider using @passthrough and @validate`);
                            }
                            new_e = new array_meta(e);
                        }
                        else {
                            new_e = array_meta(e);
                        }
                        let expected_type = typeof e;
                        let value_type = typeof new_e;
                        if (expected_type != value_type) {
                            throw new TypeError(`${this.constructor.name}.${key} array element requires type '${expected_type}', got '${value_type}' instead`);
                        }
                        ret.push(new_e);
                    }
                    return ret;
                }
                else {
                    return value;
                }
            }
            else {
                return new design_type(value);
            }
        }
        else {
            return design_type(value);
        }
    }
    set(key, value) {
        let new_value;
        let design_type = Reflect.getMetadata("design:type", this, key);
        if (Reflect.hasMetadata(__union, this, key)) {
            let values = (Reflect.getMetadata(__union, this, key));
            let is_valid = false;
            for (let i = 0; i < values.length && !is_valid; ++i) {
                if (value === values[i]) {
                    is_valid = true;
                }
            }
            if (!is_valid) {
                throw new TypeError(`${this.constructor.name}.${key} requires one of the following values:[${values}], got '${value}' instead`);
            }
        }
        if (Reflect.hasMetadata(__passthrough, this, key) && Reflect.getMetadata(__passthrough, this, key) == true) {
            new_value = value;
        }
        else {
            new_value = this.newValue(value, design_type, key);
            let expected_type = typeof new_value;
            let value_type = typeof value;
            if (expected_type != value_type) {
                throw new TypeError(`${this.constructor.name}.${key} requires type '${expected_type}', got '${value_type}' instead`);
            }
        }
        let symbols = [__code, __gt, __gte, __eq, __ne, __lt, __lte];
        for (let symbol of symbols) {
            if (Reflect.hasMetadata(symbol, this, key)) {
                let code = (Reflect.getMetadata(symbol, this, key));
                code(this, key, new_value);
            }
        }
        if (Reflect.hasMetadata(__integer, this, key)) {
            const type = typeof new_value;
            if (type != 'number') {
                throw new TypeError(`${this.constructor.name}.${key} @integer requires a numeric type, got ${type} instead`);
            }
            if (new_value != Math.floor(new_value)) {
                throw new TypeError(`${this.constructor.name}.${key} should be an integer value, got ${new_value} instead`);
            }
        }
        this[key] = new_value;
    }
    get(key) {
        return this[key];
    }
    constructor(json) {
        super();
        if (json === undefined) {
            json = {};
        }
        let json_keys = JSONObject.jsonKeys(this);
        if (!json_keys)
            return; // no properties defined
        for (let key of JSONObject.jsonKeys(this)) {
            let json_value;
            let json_key;
            // Is there a mapping this value?
            if (Reflect.hasMetadata(__mapping, this, key)) {
                json_key = Reflect.getMetadata(__mapping, this, key);
            }
            else {
                json_key = key;
            }
            json_value = json[json_key];
            // Is there a default?
            if (json_value === undefined) {
                json_value = Reflect.getMetadata(__default, this, key);
            }
            // Still undefined?
            if (json_value === undefined) {
                if (this.get(key) === undefined && !JSONObject.isOptional(this, key)) {
                    throw new TypeError(`${this.constructor.name}.${key} is required`);
                }
            }
            else {
                this.set(key, json_value);
            }
        }
    }
    static validate(code) {
        return Reflect.metadata(__code, code);
    }
    static union(values) {
        return Reflect.metadata(__union, values);
    }
    static map(newKey) {
        return Reflect.metadata(__mapping, newKey);
    }
    static isOptional(target, key) {
        let ret = Reflect.getMetadata(__optional, target, key);
        return ret == true || ret === undefined;
    }
    static jsonKeys(target) {
        return Reflect.getMetadata(__keys, target);
    }
    static preprocess(target, key) {
        let keys;
        if (Reflect.hasMetadata(__keys, target)) {
            keys = Reflect.getMetadata(__keys, target);
        }
        else {
            keys = Array();
            Reflect.defineMetadata(__keys, keys, target);
        }
        if (!keys.length || keys[keys.length - 1] != key) {
            keys.push(key);
        }
    }
    validateNumeric(key, value) {
        if (typeof value !== 'number') {
            throw new TypeError(`${this.constructor.name}.${key}: requires numeric type for comparison operators`);
        }
    }
    static array(type) {
        let meta = Reflect.metadata(__array, type);
        let ret = (target, key) => {
            JSONObject.preprocess(target, key);
            meta(target, key);
        };
        return ret;
    }
    static integer(target, key) {
        JSONObject.preprocess(target, key);
        Reflect.defineMetadata(__integer, true, target, key);
    }
    static gt(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            object.validateNumeric(key, jsonValue);
            if (!(jsonValue > capture)) {
                throw new TypeError(`${object.constructor.name}.${key}: ${jsonValue} > ${capture} requirement failed`);
            }
        };
        return Reflect.metadata(__gt, validator);
    }
    static gte(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            object.validateNumeric(key, jsonValue);
            if (!(jsonValue >= capture)) {
                throw new TypeError(`${object.constructor.name}.${key}: ${jsonValue} >= ${capture} requirement failed`);
            }
        };
        return Reflect.metadata(__gte, validator);
    }
    static lt(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            object.validateNumeric(key, jsonValue);
            if (!(jsonValue < capture)) {
                throw new TypeError(`${object.constructor.name}.${key}: ${jsonValue} < ${capture} requirement failed`);
            }
        };
        return Reflect.metadata(__lt, validator);
    }
    static lte(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            object.validateNumeric(key, jsonValue);
            if (!(jsonValue <= capture)) {
                throw new TypeError(`${object.constructor.name}.${key}: ${jsonValue} <= ${capture} requirement failed`);
            }
        };
        return Reflect.metadata(__lte, validator);
    }
    static eq(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            if (!(jsonValue == capture)) {
                throw new TypeError(`${object.constructor.name}.${key}: ${jsonValue} == ${capture} requirement failed`);
            }
        };
        return Reflect.metadata(__eq, validator);
    }
    static ne(value) {
        let capture = value;
        let validator = function (object, key, jsonValue) {
            if (!(jsonValue != capture)) {
                throw new TypeError(`${object.constructor.name}.${key}: ${jsonValue} != ${capture} requirement failed`);
            }
        };
        return Reflect.metadata(__ne, validator);
    }
    static passthrough(target, key) {
        JSONObject.preprocess(target, key);
        Reflect.defineMetadata(__passthrough, true, target, key);
    }
    static required(target, key) {
        JSONObject.preprocess(target, key);
        Reflect.defineMetadata(__optional, false, target, key);
    }
    static optional(target, key) {
        if (key === undefined) {
            let meta = Reflect.metadata(__default, target);
            let ret = (target, key) => {
                JSONObject.optional(target, key); // Make it optional
                meta(target, key);
            };
            return ret;
        }
        else {
            JSONObject.preprocess(target, key);
            Reflect.defineMetadata(__optional, true, target, key);
        }
    }
}
exports.JSONObject = JSONObject;
function validate(code) {
    return JSONObject.validate(code);
}
exports.validate = validate;
function union(values) {
    return JSONObject.union(values);
}
exports.union = union;
function required(target, key) {
    JSONObject.required(target, key);
}
exports.required = required;
function optional(target, key) {
    if (key === undefined) {
        return JSONObject.optional(target, key);
    }
    else {
        JSONObject.optional(target, key);
    }
}
exports.optional = optional;
function map(jsonKey) {
    return JSONObject.map(jsonKey);
}
exports.map = map;
function passthrough(target, key) {
    return JSONObject.passthrough(target, key);
}
exports.passthrough = passthrough;
function integer(target, key) {
    return JSONObject.integer(target, key);
}
exports.integer = integer;
function gt(value) {
    return JSONObject.gt(value);
}
exports.gt = gt;
function gte(value) {
    return JSONObject.gte(value);
}
exports.gte = gte;
function lt(value) {
    return JSONObject.lt(value);
}
exports.lt = lt;
function lte(value) {
    return JSONObject.lte(value);
}
exports.lte = lte;
function eq(value) {
    return JSONObject.eq(value);
}
exports.eq = eq;
function ne(value) {
    return JSONObject.ne(value);
}
exports.ne = ne;
function array(value) {
    return JSONObject.array(value);
}
exports.array = array;
//# sourceMappingURL=json_object.js.map